/**
 * This library names Last Java Collection Library (lastjcl).
 * 
 * Java language doesn't provide built-in methods to process effectively
 * arrays of scalar types. For example {@code byte[]}, {@code int[]}. 
 */
package org.dan.lastjcl;

import java.util.Collection;
import java.util.Iterator;

import org.dan.lastjcl.scalcol.BytesCollection;
import org.dan.lastjcl.scalist.SubList;
import org.dan.lastjcl.scaliter.BytesIteratorFactory;

/**
 * {@link IterableUtils} complements functionality of {@link Iterable} objects.
 * They get methods {@link Collection} ones have. The class is not final aim of
 * the library. It's a building brick of {@link BytesCollection}, 
 * {@link BytesIteratorFactory}, and {@link SubList} classes.
 * @author Daneel S. Yaitskov (rtfm.rtfm.rtfm@gmail.com)
 * 2012 year
 *
 */
public class IterableUtils {
	/**
	 * Analog of {@link Collection#contains(Object)}. 
	 * It's only for objects with Iterable interface.
	 * @param collection can contain {@code null}
	 * @param t can be {@code null}
	 * @return {@code true} if {@code t == null} and {@code collection} contains {@code null}
	 * or {@code t.equals(element)} returns {@code true}
	 */
	public static <T> boolean contains(Iterable<T> collection, Object t) {
		if (t == null) {
			for (T i : collection)
				if (i == null)
					return true;
			return false;
		}
		for (T i : collection)
			if (t.equals(i))
				return true;
		return false;			
	}
	
	/**
	 * Checks {@code superCollection} is superset of {@code subCollection} in other words
	 * {@code superCollection} contains all elements from {@code subCollection}. 
	 * @param superCollection can contain {@code null}
	 * @param subCollection can contain {@code null}
	 * @return {@code true} if {@code subCollection} is a subset of 
	 * {@code superCollection}. 
	 */
	public static <T> boolean containsAll(Iterable<T> superCollection, 
			Collection<?> subCollection) {
		for (Object o : subCollection) 
			if (!contains(superCollection, o))
				return false;
		return true;		
	}
	
	/** 
	 * @param collection can contain {@code null}
	 * @param T can be {@code null}
	 * @return the zero-based position of the {@code element} in the {@code collection} or
	 * {@code -1} otherwise.
	 */
	public static <T> int indexOf(Iterable<T> collection, Object element) {
		int index = 0;
		if (element == null) {
			for (T i : collection) {
				++index;
				if (i == null)
					return index;
			}
			return -1;
		}
		for (T i : collection) {
			++index;
			if (element.equals(i))
				return index;
		}
		return -1;
	}
	
	/**
	 * Consequently call .equals() of i1 item with  correspond item from o as argument.
	 * Not strict. It stops after first false result. 
	 * @param i1 cannot be null
	 * @param o can be null. any object
	 * @return true if all .equals() calls return true
	 */
	public static <T> boolean equals(Iterable<T> i1, Object o) {
		if (o == null)
			return false;
		if (o instanceof Iterable<?>) {
			Iterator<?> oi = ((Iterable<?>)o).iterator();
			Iterator<T> ii = i1.iterator();
			while (oi.hasNext() && ii.hasNext()) {
				Object o1 = oi.next();
				Object o2 = ii.next();
				if (o1 == null && o2 != null 
						|| o1 != null && o2 == null
						|| !o2.equals(o1))
					return false;
			}
			return oi.hasNext() == ii.hasNext();				
		}		
		return false;		
	}
}
