/**
 * 
 */
package org.dan.lastjcl.scalist;

import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;

import org.dan.lastjcl.IterableUtils;
import org.dan.lastjcl.ScalUtils;
import org.dan.lastjcl.scaliter.ScalIterator;

/**
 * Sub list is a projection of some other list object. It points to one continuous 
 * plot of the source list.
 * 
 * @author Daneel S. Yaitskov (rtfm.rtfm.rtfm@gmail.com)
 * 2012 year
 *
 */
public class SubList<E> implements List<E>, ScalIterator<E>, Cloneable {
	
	/** source list */
	List<E> sourceList;
	/** index of the first element in range */
	int firstIndex;
	/** range length */
	int length;
	/** the index of the next element - for Iterator interface */
	int current;
	
	/**
	 * If o is an iterable then compare correspond elements.
	 * @param o any object. 
	 */
	@Override
	public boolean equals(Object o) {
		return IterableUtils.equals(this, o);		
	}
	
	/**
	 * 
	 * @param sourceList
	 * @param firstIndex <pre>[ 0; sourceList.length - 1 ]</pre> 
	 * @param length <pre>[ 0; sourceList.length - firstIndex ]</pre>
	 * @throws IndexOutOfBoundsException 
	 */
	public SubList(List<E> sourceList, int firstIndex, int length) {
		if (firstIndex < 0 || firstIndex >= sourceList.size())
			throw new IndexOutOfBoundsException("firstIndex = " + firstIndex);
		if (length < 0 || length > sourceList.size() - firstIndex)
			throw new IndexOutOfBoundsException("length = " + length);
		
		this.sourceList = sourceList;
		this.firstIndex = firstIndex;
		this.length = length;
		this.current = 0;
	}
	
	/** {@inheritDoc} */
	public int size() {	
		return length;
	}

	/** {@inheritDoc} */
	public boolean isEmpty() {		
		return length == 0;
	}

	/**
	 * The source array is shared. For {@link Iterator} interface.
	 */
	@SuppressWarnings("unchecked")
	public Object clone() throws CloneNotSupportedException {
		SubList<E> clist = (SubList<E>)super.clone();
		clist.current = 0;
		return clist;
	}
	
	/** {@inheritDoc} */
	public boolean contains(Object o) {		
		return IterableUtils.contains(this, o);
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	public Iterator<E> iterator() {		
		try {
			return (Iterator<E>)clone();
		} catch (CloneNotSupportedException e) {
			throw new IllegalStateException(e);
		}
	}

	/** {@inheritDoc} */
	public Object[] toArray() {
		Object[] result = new Object[this.length];
		for (int i = 0; i < this.length; ++i)
			result[i] = sourceList.get(i + firstIndex);
		return result;
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	public <T> T[] toArray(T[] dstArray) {
		if (dstArray.length < this.length)
			dstArray = ScalUtils.newArray(dstArray, length);
		
		int i = 0;		
		for (; i < this.length; ++i)
			dstArray[i] = (T)sourceList.get(i+firstIndex);
		for (; i < dstArray.length; ++i)
			dstArray[i] = null;
		return dstArray;
	}

	/** 
	 * Not supported.
	 * @throws UnsupportedOperationException an array has unmodifiable structure
	 */
	public boolean add(E e) {
		throw new UnsupportedOperationException("wrapper over array");
	}

	/** 
	 * Not supported.
	 * @throws UnsupportedOperationException an array has unmodifiable structure
	 */
	public boolean remove(Object o) {
		throw new UnsupportedOperationException("wrapper over array");
	}

	/** {@inheritDoc} */
	public boolean containsAll(Collection<?> c) {
		return IterableUtils.containsAll(this, c);
	}

	/** 
	 * Not supported.
	 * @throws UnsupportedOperationException an array has unmodifiable structure
	 */
	public boolean addAll(Collection<? extends E> c) {
		throw new UnsupportedOperationException("wrapper over array");
	}

	/** 
	 * Not supported.
	 * @throws UnsupportedOperationException an array has unmodifiable structure
	 */
	public boolean addAll(int index, Collection<? extends E> c) {
		throw new UnsupportedOperationException("wrapper over array");
	}

	/** 
	 * Not supported.
	 * @throws UnsupportedOperationException an array has unmodifiable structure
	 */
	public boolean removeAll(Collection<?> c) {
		throw new UnsupportedOperationException("wrapper over array");
	}

	/** 
	 * Not supported.
	 * @throws UnsupportedOperationException an array has unmodifiable structure
	 */
	public boolean retainAll(Collection<?> c) {
		throw new UnsupportedOperationException("wrapper over array");
	}

	/** {@inheritDoc} */
	public void clear() {
		throw new UnsupportedOperationException("wrapper over array");		
	}

	/** {@inheritDoc} */
	public E get(int index) {
		checkRange(index);
		return sourceList.get(firstIndex + index);
	}

	/** {@inheritDoc} */
	public E set(int index, E element) {
		checkRange(index);
		return sourceList.set(firstIndex + index, element);
	}

	/** 
	 * Not supported.
	 * @throws UnsupportedOperationException an array has unmodifiable structure
	 */
	public void add(int index, E element) {
		throw new UnsupportedOperationException("wrapper over array");		
	}

	/** 
	 * Not supported.
	 * @throws UnsupportedOperationException an array has unmodifiable structure
	 */
	public E remove(int index) {
		throw new UnsupportedOperationException("wrapper over array");		
	}

	/** 
	 * {@inheritDoc}
	 * @param o can be {@code null}
	 */
	public int indexOf(Object o) {		
		return IterableUtils.indexOf(this, o);
	}

	/** 
	 * {@inheritDoc}
	 * @param o can be {@code null}
	 */
	public int lastIndexOf(Object o) {		
		int index = 0;		
		ScalIterator<E> si = (ScalIterator<E>)iterator();
		si.setCurrent(si.size());
		ScalListIterator<E> sli = new ScalListIterator<E>(si);
		if (o == null) {
			for (int i = length - 1; i >= 0; --i) 
				if (sli.previous() == null)
					return i;
			return -1;
		}
		for (int i = length - 1; i >= 0; --i)
			if (o.equals(sli.previous()))
				return index;		
		return -1;
	}

	/** {@inheritDoc} */
	public ListIterator<E> listIterator() {
		return new ScalListIterator<E>((ScalIterator<E>)iterator());
	}

	/** {@inheritDoc} */
	public ListIterator<E> listIterator(int index) {
		ScalIterator<E> si = (ScalIterator<E>)iterator();
		si.setCurrent(index);
		return new ScalListIterator<E>(si);		
	}

	/** {@inheritDoc} */
	public List<E> subList(int fromIndex, int toIndex) {
		return new SubList<E>(sourceList, firstIndex + fromIndex, toIndex - fromIndex);
	}
	
	/** {@inheritDoc} */
	public boolean hasNext() {		
		return current < length;
	}

	/** {@inheritDoc} */
	public E next() {				
		return get(current++);
	}

	/** 
	 * Not supported.
	 * @throws UnsupportedOperationException an array has unmodifiable structure
	 */
	public void remove() {
		throw new UnsupportedOperationException("wrapper over array");		
	}

	/**
	 * Validate {@code index} that it is in bounds of the list.
	 * @param index of list element
	 * @throws IndexOutOfBoundsException
	 */
	protected void checkRange(int index) {
		if (index >= length || index < 0)
			throw new IndexOutOfBoundsException("length = " + length);		
	}

	/** {@inheritDoc} */
	public void setCurrent(int i) {
		//checkRange(i);
		if (i > length || i < -1)
			throw new IndexOutOfBoundsException("length = " + length);
		current = i;		
	}

	/** {@inheritDoc} */
	public int getCurrent() {		
		return current;
	}

	/** {@inheritDoc} */
	public E get() {		
		return get(current);
	}

	/** {@inheritDoc} */
	public void set(E e) {
		set(current, e);		
	}
}
